@@ -2,7 +2,7 @@ from __future__ import absolute_import
 from __future__ import print_function
 
 __author__ = 'Lei Xu <eddyxu@gmail.com>'
-__version__ = '0.3.1'
+__version__ = '0.4.1'
 
 __classifiers__ = [
     'Development Status :: 3 - Alpha',
@@ -14,9 +14,10 @@ __classifiers__ = [
     'Topic :: Internet :: WWW/HTTP',
     'Topic :: Software Development :: Libraries',
     'Topic :: Software Development :: Quality Assurance',
-    'Topic :: Utilities']
+    'Topic :: Utilities',
+]
 
-__copyright__ = '2014, %s ' % __author__
+__copyright__ = '2018, %s ' % __author__
 __license__ = """
     Copyright %s.
 
@@ -33,6 +34,23 @@ __license__ = """
     limitations under the License.
     """ % __copyright__
 
+def parse_yaml_config(args):
+    """Parse yaml config"""
+    try:
+        import yaml
+    except ImportError:
+        yaml = None
+
+    yml = {}
+    try:
+        with open(args.coveralls_yaml, 'r') as fp:
+            if not yaml:
+                raise SystemExit('PyYAML is required for parsing configuration')
+            yml = yaml.load(fp)
+    except IOError:
+        pass
+    yml = yml or {}
+    return yml
 
 def run():
     """Run cpp coverage."""
@@ -46,27 +64,34 @@ def run():
     if args.verbose:
         print('encodings: {}'.format(args.encodings))
 
-    try:
-        import yaml
-    except ImportError:
-        yaml = None
-
-    yml = {}
-    try:
-        with open(args.coveralls_yaml, 'r') as fp:
-            if not yaml:
-                raise SystemExit(
-                    'PyYAML is required for parsing configuration')
-            yml = yaml.load(fp)
-    except IOError:
-        pass
-    yml = yml or {}
+    yml = parse_yaml_config(args)
 
     if not args.repo_token:
+        # try get token from yaml first
         args.repo_token = yml.get('repo_token', '')
+    if not args.repo_token:
+        # use environment COVERALLS_REPO_TOKEN as a fallback
+        args.repo_token = os.environ.get('COVERALLS_REPO_TOKEN')
+
     args.service_name = yml.get('service_name', 'travis-ci')
+
+    if not args.gcov_options:
+        args.gcov_options = yml.get('gcov_options', '')
+    if not args.root:
+        args.root = yml.get('root', '.')
+    if not args.build_root:
+        args.build_root = yml.get('build_root', '')
+
+    args.exclude.extend(yml.get('exclude', []))
+    args.include.extend(yml.get('include', []))
+    args.exclude_lines_pattern.extend(yml.get('exclude_lines_pattern', []))
+
     args.service_job_id = os.environ.get('TRAVIS_JOB_ID', '')
 
+    if args.repo_token == '' and args.service_job_id == '':
+        raise ValueError("\nno coveralls.io token specified and no travis job id found\n"
+                         "see --help for examples on how to specify a token\n")
+
     if not args.no_gcov:
         coverage.run_gcov(args)
     cov_report = coverage.collect(args)
